#!/usr/bin/env python
# -*- coding: utf-8 -*-

__author__ = "Dr. Yves-Laurent Kom Samo"
__copyright__ = "Copyright (C) 2020 KXY Technologies, Inc."
__license__ = "GPL v3.0"
__version__ = "0.0"

from kxy.api import *
from kxy.regression import *
from kxy.classification import *
from kxy.data import *