#!/usr/bin/env python
# -*- coding: utf-8 -*-

"""
This file leverages the KXY API to solve various maximum entropy rate copula problems
under concordance measures. All optimization problems are solved by the KXY API
and users require an API key that should be set in the environment variable ``KXY_API_KEY``.
"""